/* INFO ***************************************************************************
 * **                                                                            **
 * **                                   findwin                                  **
 * **                                   =======                                  **
 * **                                                                            **
 * **          Simple graphical window switcher with incremental search          **
 * **                           Version: 0.11 (20160428)                         **
 * **                                                                            **
 * **                                File: findwin.c                             **
 * **                                                                            **
 * **          For more information visit http://als.regnet.cz/findwin/          **
 * **                       Copyright (c) 2016 Aleš Koval                        **
 * **                                                                            **
 * **  This program is free software: you can redistribute it and/or modify it   **
 * **   under the terms of the GNU General Public License as published by the    **
 * **       Free Software Foundation, either version 3 of the License, or        **
 * **                    (at your option) any later version.                     **
 * **                                                                            **
 * **    This program is distributed in the hope that it will be useful, but     **
 * **         WITHOUT ANY WARRANTY; without even the implied warranty of         **
 * **            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.            **
 * **            See the GNU General Public License for more details.            **
 * **                                                                            **
 * **     You should have received a copy of the GNU General Public License      **
 * **     along with this program, most likely a file in the root directory,     **
 * **        called 'LICENSE'. If not, see <http://www.gnu.org/licenses>.        **
 * **                                                                            **
 * ************************************************************************ INFO */

/* How many icons displayed per row */
#define ICON_COLUMNS 10

#define WNCK_I_KNOW_THIS_IS_UNSTABLE
#define _GNU_SOURCE

#include <libwnck/libwnck.h>
#include <locale.h>
#include <string.h>
#include <stdlib.h>

static GtkWidget *grid, *label;

static void
draw_outline (GtkWidget *widget,
            	    cairo_t *cr,
            	    gpointer data)
{
    GtkStyleContext *context;
    GdkRGBA color;

    int selected = GPOINTER_TO_INT(g_object_get_data (G_OBJECT (widget), "selected"));

    if (selected) {
	char *name = g_object_get_data (G_OBJECT (widget), "window_name");
	gtk_label_set_markup (GTK_LABEL (label), name);

	GdkPixbuf *pixbuf = gtk_image_get_pixbuf (GTK_IMAGE (widget));
	int width = gdk_pixbuf_get_width (pixbuf);
	int height = gdk_pixbuf_get_height (pixbuf);

	/* Get rectangle color from theme */
        context = gtk_widget_get_style_context (widget);
        gtk_style_context_set_state (context, gtk_widget_get_state_flags (widget));
        gtk_style_context_lookup_color (context, "fg_color", &color);

	/* Draw rectangle around icon */
	cairo_set_line_width (cr, 2.0);
        cairo_set_source_rgb (cr, color.red, color.green, color.blue);
	cairo_rectangle (cr, 2, 2, width + 8, height + 8);
	cairo_stroke (cr);
    }
}

static GList
*get_selected_icon ()
{
    int selected;
    GList *children;

    children = gtk_container_get_children (GTK_CONTAINER (grid));
    for (; children != NULL; children = children->next) {
	selected = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (children->data), "selected"));
	if (selected)
	    return (children);
    }
    return (NULL);
}

static gboolean
key_event (GtkWidget *widget,
	    GdkEventKey *event)
{
    GList *icon, *icon_new;

    switch (event->keyval) {

	case GDK_KEY_Escape:
	    exit (0);

	case GDK_KEY_Return:
	case GDK_KEY_KP_Enter:
	    if ( (icon = get_selected_icon ())) {
    		WnckWindow *win = wnck_window_get (GPOINTER_TO_INT (g_object_get_data (G_OBJECT (icon->data), "xid")));
		WnckWorkspace *space = wnck_window_get_workspace (win);

		if (space != NULL)
		    wnck_workspace_activate (space, gtk_get_current_event_time ());

		wnck_window_activate (win, gtk_get_current_event_time ());
		exit (0);
	    }

	case GDK_KEY_Right:
	case GDK_KEY_Tab:
	    if ( (icon = get_selected_icon ())) {
		if(g_list_previous (icon)) {
		    icon_new = g_list_previous (icon)->data;

		    g_object_set_data (G_OBJECT (icon->data), "selected", GINT_TO_POINTER (0));
		    gtk_widget_queue_draw (GTK_WIDGET(icon->data));

		    g_object_set_data (G_OBJECT (icon_new), "selected", GINT_TO_POINTER (1));
		    gtk_widget_queue_draw (GTK_WIDGET (icon_new));
		}
	    }
	    return TRUE;

	case GDK_KEY_Left:
	    if ( (icon = get_selected_icon ())) {
		if(g_list_next (icon)) {
		    icon_new = g_list_next (icon)->data;

		    g_object_set_data (G_OBJECT (icon->data), "selected", GINT_TO_POINTER (0));
		    gtk_widget_queue_draw (GTK_WIDGET (icon->data));

		    g_object_set_data (G_OBJECT (icon_new), "selected", GINT_TO_POINTER (1));
		    gtk_widget_queue_draw (GTK_WIDGET (icon_new));
		}
	    }
	    return TRUE;
    }

    return FALSE;
}

static void
find_entry_changed (GtkEditable *editable,
		    gpointer user_data)
{
    WnckScreen *screen;
    GList *children, *windows;
    char *search_text, *start;
    int row = 0, column = 0;

    search_text = gtk_editable_get_chars (editable, 0, -1);

    /* Get open window list */
    screen = wnck_screen_get_default ();
    wnck_screen_force_update (screen);
    windows = wnck_screen_get_windows (screen);

    /* Clear icon grid */
    children = gtk_container_get_children (GTK_CONTAINER (grid));
    for (; children != NULL; children = children->next)
	gtk_widget_destroy (GTK_WIDGET (children->data));
    g_list_free (children);

    for (; windows != NULL; windows = windows->next) {

	WnckWindow *window = WNCK_WINDOW (windows->data);

	if (wnck_window_is_skip_pager (window) || wnck_window_is_skip_tasklist (window)) continue;

	char *window_name = g_strdup (wnck_window_get_name (window));

	if ( (start = strcasestr (window_name, search_text))) {
	    GtkWidget *icon = gtk_image_new_from_pixbuf (wnck_window_get_icon(window));

	    char *window_name_markup = g_strdup_printf ("%.*s<span color=\"red\">%s</span>%s", (int)(start - window_name), window_name, search_text, start + strlen (search_text));
	    g_object_set_data_full (G_OBJECT (icon), "window_name", window_name_markup, (GDestroyNotify) g_free);

	    /* Store window XID for later activation */
	    gint xid = wnck_window_get_xid(window);
	    g_object_set_data (G_OBJECT (icon), "xid", GINT_TO_POINTER (xid));

	    /* Outline first icon */
	    if (row == 0 && column == 0)
	        g_object_set_data (G_OBJECT (icon), "selected", GINT_TO_POINTER (1));

	    /* Add space for outline */
    	    gtk_misc_set_padding (GTK_MISC (icon), 6, 6);

	    gtk_widget_set_halign (icon, GTK_ALIGN_CENTER);
	    gtk_widget_set_valign (icon, GTK_ALIGN_CENTER);

#if GTK_MINOR_VERSION >= 12
	    gtk_widget_set_margin_start (icon, 2);
	    gtk_widget_set_margin_end (icon, 2);
#else
	    gtk_widget_set_margin_left (icon, 2);
	    gtk_widget_set_margin_right (icon, 2);
#endif
	    gtk_widget_set_margin_top (icon, 2);
	    gtk_widget_set_margin_bottom (icon, 2);

	    g_signal_connect (G_OBJECT (icon), "draw", G_CALLBACK (draw_outline), NULL);

	    gtk_grid_attach (GTK_GRID (grid), icon , column, row, 1, 1);
	    gtk_widget_show_all (grid);

	    column ++;
	    if (column >= ICON_COLUMNS) {
		column = 0;
		row ++;
	    }
	}
	g_free (window_name);

    }

    /* Any window name not match */
    if (!column)
	gtk_label_set_text (GTK_LABEL (label), "Not found.");

    g_free (search_text);
    g_list_free (windows);
}

static void
window_show (GtkWidget *window, gpointer user_data)
{
    /* Grab keyboard for popup window */
    while (gdk_keyboard_grab (gtk_widget_get_window (window), FALSE, GDK_CURRENT_TIME) != GDK_GRAB_SUCCESS) {
        sleep (0.1);
    }
}

int
main (int argc, char **argv)
{
    GtkWidget *win, *entry, *box;

    setlocale (LC_ALL, "");
    gtk_init (&argc, &argv);

    win = gtk_window_new (GTK_WINDOW_POPUP);
    gtk_window_set_position (GTK_WINDOW (win), GTK_WIN_POS_CENTER_ALWAYS);
    g_signal_connect (G_OBJECT (win), "show", G_CALLBACK (window_show), NULL);

    entry = gtk_entry_new ();
    label = gtk_label_new ("");
    gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);

    gtk_container_set_border_width (GTK_CONTAINER (win), 10);

    grid = gtk_grid_new ();

    gtk_widget_set_halign (grid, GTK_ALIGN_CENTER);
    gtk_widget_set_valign (grid, GTK_ALIGN_CENTER);

    gtk_container_set_border_width (GTK_CONTAINER (grid), 1);
    gtk_grid_set_row_spacing (GTK_GRID (grid), 3);
    gtk_grid_set_column_spacing (GTK_GRID (grid), 3);

    box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 10);
    gtk_container_add (GTK_CONTAINER (win), box);
    gtk_container_add (GTK_CONTAINER (box), entry);
    gtk_container_add (GTK_CONTAINER (box), label);
    gtk_container_add (GTK_CONTAINER (box), grid);

    g_signal_connect (win, "key_press_event", G_CALLBACK (key_event), NULL);
    g_signal_connect (entry, "changed", G_CALLBACK (find_entry_changed), NULL);

    /* Initial grid load */
    find_entry_changed (GTK_EDITABLE (entry), NULL);

    gtk_widget_show_all (win);

    gtk_main ();

    return 0;
}
